import scrapy

from ...items.base import BaseItem

class WebsiteItem(BaseItem):
    brand = scrapy.Field()
    categories_num = scrapy.Field()
    category_urls = scrapy.Field()
    total_categories = scrapy.Field()


from ..mixins.database_mixin import DatabaseMixin
from ..mixins.message_connector_mixin import MessageConnectorMixin
from ..log.default import get_default_log


class BasePipeline(DatabaseMixin, MessageConnectorMixin):

    def __init__(self):
        DatabaseMixin.__init__(self)
        MessageConnectorMixin.__init__(self)

        self.logger = None

    def get_logger(self):
        if not self.logger:
            self.logger = get_default_log()

        return self.logger


import json
import os

import pika

from .base import BaseMessageBrokerConnector


class RabbitMQConnector(BaseMessageBrokerConnector):
    def __init__(self, queue_declare, **kwargs):
        super().__init__(**kwargs)
        credentials = pika.PlainCredentials(
            os.getenv("RABBIT_MQ_USER"), os.getenv("RABBIT_MQ_PASSWORD")
        )
        parameters = pika.ConnectionParameters(
            os.getenv("RABBIT_MQ_CONN"), 5672, "/", credentials
        )
        self.rabbit_mq_conn = pika.BlockingConnection(parameters)

        self.channel = self.rabbit_mq_conn.channel()
        self.channel.queue_declare(queue=queue_declare, durable=True)

    def close(self):
        if self.rabbit_mq_conn:
            self.rabbit_mq_conn.close()

    def send(self, message, target, **kwargs):
        self.channel.basic_publish(
            exchange="",
            routing_key=target,
            body=json.dumps(message),
            properties=pika.BasicProperties(
                delivery_mode=pika.spec.PERSISTENT_DELIVERY_MODE
            ),
        )

    def receive(self, receive_from, received_callback, **kwargs):
        self.channel.basic_qos(prefetch_count=1)
        self.channel.basic_consume(
            queue=receive_from, on_message_callback=received_callback, auto_ack=False
        )

        self.channel.start_consuming()

    def send_processed_message_status(self, is_processed=True, **kwargs):
        if is_processed:
            return self.channel.basic_ack(delivery_tag=kwargs['delivery_tag'])
        else:
            return self.channel.basic_nack(delivery_tag=kwargs['delivery_tag'])




import os
import traceback

from pymongo import ReturnDocument

from ...pipelines.base import BasePipeline
from ...connectors.databases.mongodb_connector import MongoDBConnector
from ...connectors.message_brokers.rabbitmq_connector import RabbitMQConnector

class WebsitePipeline(BasePipeline):

    def __init__(self):

        self.receive_queue = None
        self.transmit_queue = None

        self.website_coll = None
        self.categories_coll = None

        super().__init__()

    def load_database_env(self):
        super().load_database_env()


    def setup_database(self, reinit=False):
        super().setup_database()
        self.website_coll = self.database[f'{os.getenv("WEBSITE_COLL")}']
        self.categories_coll = self.database[f'{os.getenv("CATEGORIES_COLL")}']

    def load_connector_env(self):
        super().load_connector_env()
        self.receive_queue = os.getenv("RECEIVE_QUEUE")
        self.connector_params['receive_queue'] = self.receive_queue
        self.transmit_queue = os.getenv("TRANSMIT_QUEUE")
        self.connector_params['transmit_queue'] = self.transmit_queue

    def init_database_connector(self):
        self.database_client = MongoDBConnector(
            connection_string=self.database_params['conn_string'],
            default_database=self.database_params['database']
        )
        self.database = self.database_client.connection[self.database_params['database']]

    def init_message_connector(self):
        self.connector_client = RabbitMQConnector()

    def process_item(self, item, spider):
        # update status of website_data
        try:
            website = None
            with self.database_client.connection.start_session() as session:
                with session.start_transaction():
                    update_obj = None

                    if item["status"] != "error":
                        update_obj = {
                            "name": item["name"],
                            "brand": item["brand"],
                            "status": item["status"],
                            "last_crawl": item["last_crawl"],
                            "css_selector": item["css_selector"],
                            "categories_num": item["categories_num"],
                            "total_categories": item["total_categories"],
                            "err_re_crawl": 0,
                            "err_re_process": 0,  # clear err when cron init
                        }
                    else:
                        website = self.website_coll.find_one({"url": item["url"]})

                        update_obj = {
                            "name": item["name"],
                            "brand": item["brand"],
                            "status": item["status"],
                            "last_crawl": item["last_crawl"],
                            "total_categories": item["total_categories"],
                            "err_re_crawl": website["err_re_crawl"] + 1,
                            "err_re_process": 0,  # clear err when cron init
                        }

                    website = self.website_coll.find_one_and_update(
                        {"url": item["url"]},
                        {
                            "$set": update_obj,
                        },
                        return_document=ReturnDocument.AFTER,
                        session=session,
                    )

            # send all categories_url to categories_queue

            if website:
                for category in item["category_urls"]:
                    # test rabbit
                    if category != item["category_urls"][0]:
                        continue

                    body_message = {
                        "website_id": str(website["_id"]),
                        "website_url": website["url"],
                        "website_name": website["name"],
                        "url": category,
                        "spider_name": website["spider_name"],
                    }

                    # send message to rabbitmq (categories_queue)
                    self.send_message(body_message, self.connector_params['transmit_queue'])

            self.get_logger().info(
                f'[*] Crawling website {item["url"]} completed at {item["last_crawl"]}: status - "{item["status"]}"'
            )
        except Exception as error:
            # traceback.print_exc()
            # + log to log server
            self.get_logger().error(
                f'[x] Error when process data crawled in pipeline website (url: {item["url"]}). \n'
                + traceback.format_exc()
            )

        return item

    def close_spider(self, spider):
        # disconnect database
        self.database_client.close()

        # disconnect rabbitmq
        self.connector_client.close()


from .base import BaseSpider
from ..items.product_item import ProductItem


class ProductSpider(BaseSpider):
    instance = "product"
    allProduct = []
    tempProduct = []

    def __init__(self, url):
        super().__init__(url)

    @staticmethod
    def get_new_item_instance():
        return ProductItem()

    def set_default_value(self):
        items = super().set_default_value()
        items["product_id"] = ""  # update when error
        items["competitor"] = ""  # update when error
        items["good_nm"] = ""  # update when error
        items["normal_price"] = None  # update when error
        items["sale_price"] = None  # update when error
        items["shipping_price"] = 0  # update when error
        items["shipping_type"] = 1  # update when error
        items["shipping_condition"] = 0  # update when error
        items["options"] = []  # update when error
        items["total_options"] = 0  # update when error
        items["currency"] = "KRW"  # update when error
        return items



import pymongo

from .base import BaseDatabaseConnector

class MongoDBConnector(BaseDatabaseConnector):

    def __init__(self, connection_string, default_database=None):
        super().__init__(connection_string)
        self.database = self.connection[default_database]

    @staticmethod
    def get_connection(**kwargs):
        return pymongo.MongoClient(kwargs['connection_string'])

    def close(self):
        if self.connection:
            self.connection.close()

    def get(self, key, obj_type, **kwargs):
        return self.database[obj_type].find_one(key)

    def get_list(self, criteria, obj_type, **kwargs):
        return self.database[obj_type].find()

    def insert(self, obj, obj_type, **kwargs):
        return self.database[obj_type].insert_one(obj)

    def update(self, criteria, new_obj, obj_type, **kwargs):
        return self.database[obj_type].update_one(
            criteria,
            {
                "$set": new_obj
            }
        )

